#pragma once

#include <type_traits>

#include <dune/common/typetraits.hh>
#include <dune/common/std/optional.hh>
#include <dune/geometry/type.hh>

namespace AMDiS
{
  namespace Impl
  {
    template <class E, class = Dune::void_t<>>
    struct ContextTypes
    {
      using Entity = E;
      using LocalGeometry = typename E::Geometry;
    };

    // specialization for intersections
    template <class I>
    struct ContextTypes<I, Dune::void_t<decltype(std::declval<I>().inside())>>
    {
      using Entity = typename I::Entity;
      using LocalGeometry = typename I::LocalGeometry;
    };

  } // end namespace Impl


  /// \brief Wrapper class for element and geometry
  /**
   * A LocalContext can be either a grid entity of codim 0 (called an element)
   * or an intersection of elements. The element and its geometry may be stored
   * externally and can be passed along with the localContext object.
   * Since an intersection has a geometry (and localGeometry) different from the
   * geometry (and localGeometry) of the entity it belongs to, these objects
   * are provided as well.
   **/
  template <class LocalContextType>
  struct ContextGeometry
  {
  public:

    using LocalContext = LocalContextType;
    using Element = typename Impl::ContextTypes<LocalContext>::Entity;
    using Geometry = typename Element::Geometry;
    using LocalGeometry = typename Impl::ContextTypes<LocalContext>::LocalGeometry;

    using IsEntity = std::is_same<Element, LocalContext>;

    enum {
      dim = Geometry::mydimension,    //< the dimension of the grid element
      dow = Geometry::coorddimension  //< the dimension of the world
    };

    /// Constructor. Stores pointer to localContext, element, and geometry.
    ContextGeometry(LocalContext const& localContext, Element const& element, Geometry const& geometry)
      : localContext_(&localContext)
      , element_(&element)
      , geometry_(&geometry)
    {}

  public:

    Element const& element() const
    {
      return *element_;
    }

    LocalContext const& localContext() const
    {
      return *localContext_;
    }

    Geometry const& geometry() const
    {
      return *geometry_;
    }

    LocalGeometry const& localGeometry() const
    {
      return localGeometryImpl(IsEntity{});
    }


  public:

    /// Coordinate `p` given in `localGeometry`, transformed to coordinate in `geometry`.
    template <class Coordinate>
    decltype(auto) local(Coordinate const& p) const
    {
      return localImpl(p, IsEntity{});
    }

    /// Transformation of coordinate `p` given in `localGeometry` to world space coordinates.
    template <class Coordinate>
    decltype(auto) global(Coordinate const& p) const
    {
      return geometry_->global(p);
    }

    /// Return the geometry-type of the localContext
    Dune::GeometryType type() const
    {
      return localContext_->type();
    }

    /// The integration element from the `localGeometry`, the quadrature points are
    /// defined in.
    template <class Coordinate>
    auto integrationElement(Coordinate const& p) const
    {
      return localGeometry().integrationElement(p);
    }

  private: // implementation detail

    // position for elements
    template <class Coordinate>
    Coordinate const& localImpl(Coordinate const& p, std::true_type) const
    {
      return p;
    }

    // position for intersection
    template <class Coordinate>
    auto localImpl(Coordinate const& p, std::false_type) const
    {
      return localGeometry().global(p);
    }

    // local-geometry is the same as geometry
    Geometry const& localGeometryImpl(std::true_type) const
    {
      return *geometry_;
    }

    // local-geometry of intersection in inside element
    LocalGeometry const& localGeometryImpl(std::false_type) const
    {
      if (!localGeometry_)
        localGeometry_.emplace(localContext_->geometryInInside());

      return *localGeometry_;
    }

  private:
    LocalContext const* localContext_;
    Element const* element_;
    Geometry const* geometry_;

    // The localGeometry may be constructed only if needed
    mutable Dune::Std::optional<LocalGeometry> localGeometry_;
  };

} // end namespace AMDiS
