#pragma once

#include <cassert>
#include <type_traits>

#include <amdis/GridFunctions.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/QuadratureFactory.hpp>
#include <amdis/common/Transposed.hpp>
#include <amdis/common/Utility.hpp>
#include <amdis/utility/FiniteElementType.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  /// \brief The main implementation of an CRTP-base class for operators using a grid-function
  /// coefficient to be used in a \ref LocalAssembler.
  /**
   * An Operator that takes a GridFunction as coefficient.
   * Provides quadrature rules and handles the evaluation of the GridFunction at
   * local coordinates.
   *
   * The class is specialized, by deriving from it, in \ref GridFunctionOperator.
   *
   * \tparam Derived      The class derived from GridFunctionOperatorBase
   * \tparam LocalContext The Element or Intersection type
   * \tparam GridFunction The GridFunction, a LocalFunction is created from, and
   *                      that is evaluated at quadrature points.
   *
   * **Requirements:**
   * - `LocalContext` models either Entity (of codim 0) or Intersection
   * - `GridFunction` models the \ref Concepts::GridFunction
   **/
  template <class Derived, class LocalContext, class GridFunction>
  class GridFunctionOperatorBase
      : public LocalOperator<Derived, LocalContext>
  {
    /// The type of the localFunction associated with the GridFunction
    using LocalFunction = decltype(localFunction(std::declval<GridFunction>()));

    /// The Codim=0 entity of the grid, the localFunction can be bound to
    using Element = typename Impl::ContextTypes<LocalContext>::Entity;
    /// The geometry-type of the grid element
    using Geometry = typename Element::Geometry;

    /// The type of the local coordinates in the \ref Element
    using LocalCoordinate = typename GridFunction::EntitySet::LocalCoordinate;

    /// A factory for QuadratureRules that incooperate the order of the LocalFunction
    using QuadFactory = QuadratureFactory<typename Geometry::ctype, LocalContext::mydimension, LocalFunction>;

  public:
    /// \brief Constructor. Stores a copy of `gridFct`.
    /**
     * A GridFunctionOperator takes a gridFunction and the
     * differentiation order of the operator, to calculate the
     * quadrature degree in \ref getDegree.
     **/
    GridFunctionOperatorBase(GridFunction const& gridFct, int termOrder)
      : gridFct_(gridFct)
      , termOrder_(termOrder)
    {}

    /// \brief Binds operator to `element` and `geometry`.
    /**
     * Binding an operator to the currently visited element in grid traversal.
     * Since all operators need geometry information, the `Element::Geometry`
     * object `geometry` is created once, during grid traversal, and passed in.
     *
     * By default, it binds the \ref localFct_ to the `element` and the Quadrature
     * factory to the localFunction.
     **/
    void bind_impl(Element const& element, Geometry const& geometry)
    {
      assert( bool(quadFactory_) );
      localFct_.emplace(localFunction(gridFct_));
      localFct_->bind(element);
      quadFactory_->bind(localFct_.value());
    }

    /// Unbinds operator from element.
    void unbind_impl()
    {
      localFct_->unbind();
      localFct_.reset();
    }

    /// Create a quadrature factory from a PreQuadratureFactory, e.g. class derived from \ref QuadratureFactory
    template <class PreQuadFactory>
    void setQuadFactory(PreQuadFactory const& pre)
    {
      using ctype = typename Geometry::ctype;
      quadFactory_.emplace(makeQuadratureFactory<ctype, LocalContext::mydimension, LocalFunction>(pre));
    }

  protected:
    /// Return expression value at LocalCoordinates
    auto coefficient(LocalCoordinate const& local) const
    {
      assert( this->bound_ );
      return (*localFct_)(local);
    }

    /// Create a quadrature rule using the \ref QuadratureFactory by calculating the
    /// quadrature order necessary to integrate the (bi)linear-form accurately.
    template <class... Nodes>
    auto const& getQuadratureRule(Dune::GeometryType type, Nodes const&... nodes) const
    {
      assert( bool(quadFactory_) );
      int quadDegree = this->getDegree(termOrder_, quadFactory_->order(), nodes...);
      return quadFactory_->rule(type, quadDegree);
    }

  private:
    /// The gridFunction to be used within the operator
    GridFunction gridFct_;

    /// localFunction associated with gridFunction. Mus be updated whenever gridFunction changes.
    Dune::Std::optional<LocalFunction> localFct_;

    /// Assign each element type a quadrature rule
    Dune::Std::optional<QuadFactory> quadFactory_;

    /// the derivative order of this operator (in {0, 1, 2})
    int termOrder_ = 0;
  };


  /// \brief The transposed operator, implemented in term of its transposed by
  /// calling \ref getElementMatrix with inverted arguments.
  template <class Derived, class Transposed>
  class GridFunctionOperatorTransposed
      : public LocalOperator<Derived, typename Transposed::LocalContext>
  {
    template <class T, class... Args>
    using Constructable = decltype( new T(std::declval<Args>()...) );

  public:
    template <class... Args,
      std::enable_if_t<Dune::Std::is_detected<Constructable, Transposed, Args...>::value, int> = 0>
    GridFunctionOperatorTransposed(Args&&... args)
      : transposedOp_(std::forward<Args>(args)...)
    {}

    /// Redirects the bind call top the transposed operator
    template <class Element, class Geometry>
    void bind_impl(Element const& element, Geometry const& geometry)
    {
      transposedOp_.bind_impl(element, geometry);
    }

    /// Redirects the unbind call top the transposed operator
    void unbind_impl()
    {
      transposedOp_.unbind_impl();
    }

    /// Redirects the setQuadFactory call top the transposed operator
    template <class PreQuadFactory>
    void setQuadFactory(PreQuadFactory const& pre)
    {
      transposedOp_.setQuadFactory(pre);
    }

    /// Apply the assembling to the transposed elementMatrix with interchanged row-/colNode
    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      auto elementMatrixTransposed = transposed(elementMatrix);
      transposedOp_.getElementMatrix(
        context, colNode, rowNode, elementMatrixTransposed);
    }

  private:
    Transposed transposedOp_;
  };

  template <class Tag, class PreGridFct, class PreQuadFactory>
  struct PreGridFunctionOperator
  {
    Tag tag;
    PreGridFct expr;
    PreQuadFactory quadFactory;
  };

  /// Store tag and expression to create a \ref GridFunctionOperator
  template <class Tag, class PreGridFct, class... QuadratureArgs>
  auto makeOperator(Tag tag, PreGridFct const& expr, QuadratureArgs&&... args)
  {
    auto preQuadFactory = makePreQuadratureFactory(std::forward<QuadratureArgs>(args)...);
    return PreGridFunctionOperator<Tag, PreGridFct, decltype(preQuadFactory)>{tag, expr, preQuadFactory};
  }

  /** @} **/

#ifndef DOXYGEN

  /// The base-template for GridFunctionOperators
  /**
   * An operator can specialize this class, by deriving from \ref GridFunctionOperatorBase.
   * With the generic function \ref makeLocalOperator, an instance is created. To
   * distinguisch different GridFunction operators, a tag can be provided that has no
   * other effect.
   *
   * \tparam Tag  An Identifier for this GridFunctionOperator
   * \tparam LocalContext  An Element or Intersection the operator is evaluated on
   * \tparam GridFct  A GridFunction evaluated in local coordinates on the bound element
   **/
  template <class Tag, class LocalContext, class GridFct>
  class GridFunctionOperator
      : public GridFunctionOperatorBase<GridFunctionOperator<Tag, LocalContext, GridFct>, LocalContext, GridFct>
  {};


  /// Generate an \ref GridFunctionOperator from a PreOperator (tag, expr).
  /// @{
  template <class LocalContext, class Tag, class... Args, class GridView>
  auto makeLocalOperator(PreGridFunctionOperator<Tag, Args...> const& op, GridView const& gridView)
  {
    auto gridFct = makeGridFunction(op.expr, gridView);
    using GridFctOp = GridFunctionOperator<Tag, LocalContext, decltype(gridFct)>;
    GridFctOp localOperator{op.tag, gridFct};
    localOperator.setQuadFactory(op.quadFactory);
    return localOperator;
  }

  template <class LocalContext, class Tag, class... Args, class GridView>
  auto makeLocalOperator(std::reference_wrapper<PreGridFunctionOperator<Tag, Args...>> op, GridView const& gridView)
  {
    PreGridFunctionOperator<Tag, Args...> const& op_ref = op;
    auto gridFct = makeGridFunction(std::ref(op_ref.expr), gridView);
    using GridFctOp = GridFunctionOperator<Tag, LocalContext, decltype(gridFct)>;
    GridFctOp localOperator{op_ref.tag, gridFct};
    localOperator.setQuadFactory(op_ref.quadFactory);
    return localOperator;
  }
  /// @}


  /// Generate a shared_ptr to \ref GridFunctionOperator from a PreOperator (tag, expr).
  /// @{
  template <class LocalContext, class Tag, class... Args, class GridView>
  auto makeLocalOperatorPtr(PreGridFunctionOperator<Tag, Args...> const& op, GridView const& gridView)
  {
    auto gridFct = makeGridFunction(op.expr, gridView);
    using GridFctOp = GridFunctionOperator<Tag, LocalContext, decltype(gridFct)>;
    auto localOperator = std::make_shared<GridFctOp>(op.tag, gridFct);
    localOperator->setQuadFactory(op.quadFactory);
    return localOperator;
  }

  template <class LocalContext, class Tag, class... Args, class GridView>
  auto makeLocalOperatorPtr(std::reference_wrapper<PreGridFunctionOperator<Tag, Args...>> op, GridView const& gridView)
  {
    PreGridFunctionOperator<Tag, Args...> const& op_ref = op;
    auto gridFct = makeGridFunction(std::ref(op_ref.expr), gridView);
    using GridFctOp = GridFunctionOperator<Tag, LocalContext, decltype(gridFct)>;
    auto localOperator = std::make_shared<GridFctOp>(op_ref.tag, gridFct);
    localOperator->setQuadFactory(op_ref.quadFactory);
    return localOperator;
  }
  /// @}

#endif // DOXYGEN

} // end namespace AMDiS
